package gbaMus;

import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

/**
 * GBA SongRiper (c) 2012 by Bregalad
 * This is free and open source software
 * 
 * This program convert a GBA song for the sappy sound engine into MIDI (.mid) format.
 */

public class SongRiper {

	static DataOutputStream outMID;
	static RandomAccessFile inGBA;

	static int base_address;
	static int[] track_ptr;
	static int[] last_cmd;
	static int[] last_key;
	static int[] last_vel;
	static int[] counter;
	static int[] return_ptr;
	static int[] key_shift;
	static boolean[] return_flag;
	static boolean[] track_completed;
	static boolean end_flag;
	static boolean loop_flag;
	static int loop_adr;

	static int[] lfo_delay_ctr;
	static int[] lfo_delay;
	static int[] lfo_depth;	static int[] lfo_type;
	static boolean[] lfo_flag;

	static int simultaneous_notes_ctr;
	static int simultaneous_notes_max;

	static ArrayList<Note> notes_playing;
	
	static int bank_number;
	static boolean bank_used = false;
	static boolean rc = false;
	static boolean gs = false;
	static boolean xg = false;
	static boolean lv = false;
	static boolean sv = false;

	static MIDI midi;

	public static void main(String[] args) throws Exception {
		System.out.println("GBA ROM sequence ripper (c) 2012 Bregalad");

		parseArguments(args);

		inGBA.seek(base_address);

		int track_amnt = inGBA.read();
		if(track_amnt < 1 || track_amnt > 16) {
			System.out.println("Invalid amount of tracks ! (must be 1-16)");
			throw new Exception();
		} else
			System.out.println(track_amnt + " tracks.");


		//Open output file once we know the pointer points to correct data
		//(this avoids creating blank files when there is an error)
		try {
			outMID = new DataOutputStream(new FileOutputStream(args[1]));
		} catch (FileNotFoundException e) {
			System.out.println("Invalid output file : " + args[1]);
			System.exit(-1);
		}

		System.out.println("Converting...");

		midi = new MIDI(24);
		simultaneous_notes_max = 0;
		simultaneous_notes_ctr = 0;
		end_flag = false;
		loop_flag = false;

		if(rc) {
			//Make the drum channel last in the list, hopefully reducing the risk of it being used
			midi.chanel_reorder = new int[]{0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 13, 14, 15, 9};
		}

		if(gs) {
			//GS reset
			midi.addSysex(new byte[]{0x41, 0x10, 0x42, 0x12, 0x40, 0x00, 0x7f, 0x00, 0x41});
			//Part 10 to normal
			midi.addSysex(new byte[]{0x41, 0x10, 0x42, 0x12, 0x40, 0x10, 0x15, 0x00, 0x1b});
		}

		if(xg) {
			//XG reset
			midi.addSysex(new byte[]{0x43, 0x10, 0x4C, 0x00, 0x00, 0x7E, 0x00});
		}

		midi.addMarker("Converted by SequenceRipper");

		track_ptr = new int[track_amnt];
		last_cmd = new int[track_amnt];
		last_key = new int[track_amnt];
		last_vel = new int[track_amnt];
		counter = new int[track_amnt];
		return_ptr = new int[track_amnt];
		return_flag = new boolean[track_amnt];
		track_completed = new boolean[track_amnt];
		key_shift = new int[track_amnt];
		notes_playing = new ArrayList<Note>();

		lfo_delay = new int[track_amnt];
		lfo_delay_ctr = new int[track_amnt];
		lfo_depth = new int[track_amnt];
		lfo_flag = new boolean[track_amnt];
		lfo_type = new int[track_amnt];

		byte unknown = inGBA.readByte();
		byte priority = inGBA.readByte();
		byte reverb = inGBA.readByte();

		int intr_bank_address = GBAMusRiper.getPointer(inGBA.readInt());

		//Read table of pointers
		for(int i=0; i<track_amnt; i++) {
			track_ptr[i] = GBAMusRiper.getPointer(inGBA.readInt());

			lfo_depth[i] = 0;
			lfo_delay[i] = 0;
			lfo_flag[i] = false;

			if(reverb < 0)
				//add reverb controller on all tracks
				midi.addController(i, 91, lv ? (int)Math.sqrt((reverb&0x7f)*127.0) : reverb&0x7f);
		}

		//Search for loop address of track #0
		if(track_amnt > 1)
			//If 2 or more track, end of track is before start of track 2
			inGBA.seek(track_ptr[1] - 9);
		else
			//If only a single track, the end is before start of header data
			inGBA.seek(base_address - 9);

		//Read where in track 1 the loop starts
		for(int i=0; i<5; i++)
			if(inGBA.read() == 0xb2) {
				loop_flag = true;
				loop_adr = GBAMusRiper.getPointer(inGBA.readInt());
				break;
			}

		//This is the main loop which will process all channels
		//until they are all inactive
		int i = 100000;
		while(tick(track_amnt)) {
			if(i-- == 0) {
				System.out.println("Time out");
				break;
			}
		}

		//If a loop was detected this is its end
		if(loop_flag)
			midi.addMarker("loopEnd");

		System.out.println("Maximum simultaneous notes : " + simultaneous_notes_max);

		System.out.print("Dump complete. Now outputing MIDI file...");
		midi.write(outMID);
		//Close files
		inGBA.close();
		System.out.println(" Done !");
	}

	static boolean tick(int track_amnt)  throws IOException {
		//Tick all playing notes, and remove notes which
		//have been keyed off OR which are infinite length from the list
		for(int i=0; i<notes_playing.size(); i++)
			if(notes_playing.get(i).tick() || notes_playing.get(i).counter < 0)
				notes_playing.remove(i--);

		//Process all tracks
		for(int track = 0; track<track_amnt; track++) {
			counter[track]--;
			//Process events until counter non-null or pointer null
			//This might not be executed if counter both are non null.
			while(track_ptr[track] != 0 && !end_flag && counter[track] <= 0) {
				//Check if we're at loop start point
				if(track == 0 && loop_flag && !return_flag[0] && !track_completed[0] && track_ptr[0] == loop_adr)
					midi.addMarker("loopStart");

				process_event(track);
			}
		}

		for(int track = 0; track<track_amnt; track++) {
			process_lfo(track);
		}

		//Compute if all still active channels are completely decoded
		boolean all_completed_flag = true;
		for(int i = 0; i < track_amnt; i++)
			all_completed_flag &= track_completed[i];

		//If everything is completed, the main program should quit its loop
		if(all_completed_flag)
			return false;

		//Make note on events for this tick
		//(it's important they are made after all other events)
		for(int i=0; i<notes_playing.size(); i++)
			notes_playing.get(i).makeNoteOnEvent();

		//Increment MIDI time
		midi.time_ctr++;

		return true;
	}

	static void addSimultaneousNote() {
		//Update simultaneous notes max.
		if(++simultaneous_notes_ctr > simultaneous_notes_max)
			simultaneous_notes_max = simultaneous_notes_ctr;
	}

	//LFO logic on tick
	static void process_lfo(int track) {
		if(sv && lfo_delay_ctr[track] != 0) {
			//Decrease counter if it's value was nonzero
			if(--lfo_delay_ctr[track] ==0) {
				//If 1->0 transition we need to add a signal to start the LFO
				if(lfo_type[track]==0)
					//Send a controller 1 if pitch LFO
					midi.addController(track, 1, (lfo_depth[track] < 16) ? lfo_depth[track] * 8 : 127);
				else
					//Send a channel aftertouch otherwise
					midi.addChanAft(track, (lfo_depth[track] < 16) ? lfo_depth[track] * 8 : 127);
				lfo_flag[track] = true;
			}
		}
	}

	static void start_lfo(int track) {
		//Reset down delay counter to its initial value
		if(sv && lfo_delay[track] != 0)
			lfo_delay_ctr[track] = lfo_delay[track];
	}

	static void stop_lfo(int track) {
		//Cancel a LFO if it was playing,
		if(sv && lfo_flag[track]) {
			if(lfo_type[track]==0)
				midi.addController(track, 1, 0);
			else
				midi.addChanAft(track, 0);
			lfo_flag[track] = false;
		} else
			lfo_delay_ctr[track] = 0;			//cancel delay counter if it wasn't playing
	}

	static void process_event(int track) throws IOException {
		//Length table for notes and rests
		final int[] lenTbl = {
			0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
			16, 17, 18, 19, 20, 21, 22, 23, 24, 28, 30, 32, 36,
			40, 42, 44, 48, 52, 54, 56, 60, 64, 66, 68, 72, 76, 78,
			80, 84, 88, 90, 92, 96
		};

		inGBA.seek(track_ptr[track]);
		//Read command
		int command = inGBA.read();
		track_ptr[track]++;
		int arg1;
		//Repeat last command, the byte read was in fact the first argument
		if(command < 0x80) {
			arg1 = command;
			command = last_cmd[track];
		}

		//Delta time command
		else if(command <= 0xb0) {
			counter[track] = lenTbl[command - 0x80];
			return;
		}

		//End track command
		else if(command == 0xb1) {
			//Null pointer
			track_ptr[track] = 0;
			track_completed[track] = true;
			return;
		}

		//Jump command
		else if(command == 0xb2)  {
			track_ptr[track] = GBAMusRiper.getPointer(inGBA.readInt());

			//detect the end track
			track_completed[track] = true;
			return;
		}

		//Call command
		else if(command == 0xb3) {
			int addr = GBAMusRiper.getPointer(inGBA.readInt());
	
			//Return address for the track
			return_ptr[track] = track_ptr[track] + 4;
			//Now points to called address
			track_ptr[track] = addr;
			return_flag[track] = true;
			return;
		}

		//Return command
		else if(command == 0xb4) {
			if(return_flag[track]) {
				track_ptr[track] = return_ptr[track];
				return_flag[track] = false;
			}
			return;
		}

		//Tempo change
		else if(command == 0xbb) {
			int tempo = 2*inGBA.read();
			track_ptr[track]++;
			midi.addTempoChange(tempo);
			return;
		}

		else {
			//Normal command
			last_cmd[track] = command;
			//Need argument
			arg1 = inGBA.read();
			track_ptr[track]++;
		}

		//Note on with specified length command
		if(command >= 0xd0) {
			int key, vel;
			int len_ofs = 0;
			//Is arg1 a key value ?
			if(arg1 < 0x80) {
				//Yes -> use new key value
				key = arg1;
				last_key[track] = key;

				int arg2 = inGBA.read();
				//Is arg2 a velocity ?
				if(arg2 < 0x80) {
					//Yes -> use new velocity value
					vel = arg2;
					last_vel[track] = vel;
					track_ptr[track]++;

					int arg3 = inGBA.read();

					//Is there a length offset ?
					if(arg3 < 0x80) {
						//Yes -> read it and increment pointer
						len_ofs = arg3;
						track_ptr[track]++;
					}
					
				} else {
					//No -> use previous velocity value
					vel = last_vel[track];
				}
			} else {
				//No -> use last value
				key = last_key[track];
				vel = last_vel[track];
				track_ptr[track]--;		//Seek back, as arg 1 is unused and belong to next event !
			}

			//Linearize velocity if needed
			if(lv)
				vel = (int)Math.sqrt(127.0 * vel);

			Note n = new Note(midi, track, lenTbl[command - 0xd0 + 1] + len_ofs, key + key_shift[track], vel);
			notes_playing.add(n);

			return;
		}

		//Other commands
		switch(command) {

		//Key shift
		case 0xbc : {
			key_shift[track] = arg1;
		}break;

		//Set instrument
		case 0xbd : {
			if(bank_used)
				if(!xg) {
					midi.addController(track, 0, bank_number);
				} else {
					midi.addController(track, 0, bank_number >> 7);
					midi.addController(track, 32, bank_number & 0x7f);
				}
			midi.addPChange(track, arg1);
		}break;

		//Set volume
		case 0xbe : {
			//Linearize volume if needed
			int volume = lv ? (int)Math.sqrt(127.0 * arg1) : arg1;
			midi.addController(track, 7, volume);
		}break;

		//Set panning
		case 0xbf : {
			midi.addController(track, 10, arg1);
		}break;

		//Pitch bend
		case 0xc0 : {
			midi.addPitchBend(track, (byte)arg1);
		}break;

		//Pitch bend range
		case 0xc1 : {
			if(sv)
				midi.addRPN(track, 0, (byte)arg1);
			else
				midi.addController(track, 20, arg1);
		}break;

		//LFO Speed
		case 0xc2 : {
			midi.addController(track, 21, arg1);
		}break;

		//LFO delay
		case 0xc3 : {
			if(sv)
				lfo_delay[track] = arg1;
			else
				midi.addController(track, 26, arg1);
		}break;

		//LFO depth
		case 0xc4 : {
			if(sv) {
				if(lfo_delay[track] == 0) {
					if(lfo_type[track]==0)
						midi.addController(track, 1, arg1>12 ? 127 : 10 * arg1);
					else
						midi.addChanAft(track, arg1>12 ? 127 : 10 * arg1);

					lfo_flag[track] = true;
				}
				lfo_depth[track] = arg1;
			}
			else
				midi.addController(track, 1, arg1);
		}break;

		//LFO type
		case 0xc5 : {
			if(sv)
				lfo_type[track] = arg1;
			else
				midi.addController(track, 22, arg1);
		}break;

		//Detune
		case 0xc8 : {
			if(sv)
				midi.addRPN(track, 1, (byte)arg1);
			else
				midi.addController(track, 24, arg1);
		}break;

		//Key off
		case 0xce : {
			int key, vel = 0;

			//Is arg1 a key value ?
			if(arg1 < 0x80) {
				//Yes -> use new key value
				key = arg1;
				last_key[track] = key;
			} else {
				//No -> use last value
				key = last_key[track];
				vel = last_vel[track];
				track_ptr[track]--;		//Seek back, as arg 1 is unused and belong to next event !
			}

			midi.addNoteOff(track, key + key_shift[track], vel);
			stop_lfo(track);
			simultaneous_notes_ctr --;
		}break;

		//Key on
		case 0xcf : {
			int key, vel;
			//Is arg1 a key value ?
			if(arg1 < 0x80) {
				//Yes -> use new key value
				key = arg1;
				last_key[track] = key;

				int arg2 = inGBA.read();
				//Is arg2 a velocity ?
				if(arg2 < 0x80) {
					//Yes -> use new velocity value
					vel = arg2;
					last_vel[track] = vel;
					track_ptr[track]++;					
				} else {
					//No -> use previous velocity value
					vel = last_vel[track];
				}
			} else {
				//No -> use last value
				key = last_key[track];
				vel = last_vel[track];
				track_ptr[track]--;		//Seek back, as arg 1 is unused and belong to next event !
			}
			//Linearize velocity if needed
			if(lv)
				vel = (int)Math.sqrt(127.0 * vel);

			//Make note of infinite length
			Note n = new Note(midi, track, -1, key + key_shift[track], vel);
			notes_playing.add(n);
		}break;

		default :
			System.out.println("Unknown command : 0x" + Integer.toHexString(command));
		}
	}

	static void parseArguments(String[] args) {
		if (args.length < 3)
			printInstructions();

		//Open the input and output files
		try{
			inGBA = new RandomAccessFile(args[0], "r");
		}catch(FileNotFoundException e) {
			System.out.println("Invalid input GBA file : " + args[0]);
			System.exit(-1);
		}

		//Determine if base address is given in decimal or hexadecimal
		if(args[2].substring(0,2).equals("0x"))
			base_address = Integer.parseInt(args[2].substring(2), 16);
		else
			base_address = Integer.parseInt(args[2]);

		for(int i=3; i<args.length; i++) {
			if(args[i].substring(0,2).equals("-b")) {
				if(args[i].length()<3) printInstructions();
				bank_number = Integer.parseInt(args[i].substring(2));
				bank_used = true;
			}
			else if(args[i].equals("-rc"))
				rc = true;
			else if(args[i].equals("-gs"))
				gs = true;
			else if(args[i].equals("-xg"))
				xg = true;
			else if(args[i].equals("-lv"))
				lv = true;
			else if(args[i].equals("-sv"))
				sv = true;
			else
				printInstructions();
		}
	}

	static void printInstructions() {
		System.out.println("Rips sequence data from a GBA game using sappy sound engine to MIDI (.mid) format.");
		System.out.println("\nUsage : MidiRipper infile.gba outfile.mid song_adress [-b1 -gm -gs -xg]");
		System.out.println("-b : Bank : forces all patches to be in the specified bank (0-127)");
		System.out.println("In general MIDI, midi channel 10 is reserved for drums");
		System.out.println("Unfortunatly, we do not want to use any \"drums\" in the output file");
		System.out.println("I have 3 modes to fix this problem");
		System.out.println("-rc : Rearrange Channels. This will avoid using the channel 10, and use it at last ressort only if all 16 channels should be used");
		System.out.println("-gs : This will send a GS system exculsive message to tell the player channel 10 is not drums");
		System.out.println("-xg : This will send a XG system exclusive message, and force banks number which will disable \"drums\"");
		System.out.println("-lv : Linearise volume and velocities. This should be used to have the output \"sound\" like the original song," +
							"but shouldn't be used to get an exact dump of sequence data.");
		System.out.println("-sv : Simulate vibrato. This will insert controllers in real time to simulate a vibrato, instead of just" +
							"when commands are given. Like -lv, this should be used to have the output \"sound\" like the original song," +
							"but shouldn't be used to get an exact dump of sequence data.");
		System.out.println("It is possible, but not recommanded, to use more than one of these flags at a time.");
		System.exit(-1);
	}
}

//Note class
//this was needed to properly handle polyphony on all channels...

class Note {
	MIDI midi;
	int counter;
	int key;
	int vel;
	int chn;
	boolean event_made;

	//Create note and key on event
	Note(MIDI midi, int chn, int len, int key, int vel) {
		this.midi = midi;
		this.counter = len;
		this.key = key;
		this.vel = vel;
		this.chn = chn;
		event_made = false;

		SongRiper.start_lfo(chn);

		SongRiper.addSimultaneousNote();
	}

	void makeNoteOnEvent() {
		if(!event_made) {
			midi.addNoteOn(chn, key, vel);
			event_made = true;
		}
	}

	//Tick counter, if it becomes zero
	//then create key off event
	//this function returns "true" when the note should be freed from memory
	boolean tick() {
		if(counter > 0 && --counter == 0) {
			midi.addNoteOff(chn, key, vel);
			SongRiper.stop_lfo(chn);
			SongRiper.simultaneous_notes_ctr --;
			return true;
		}
		else
			return false;
	}
}
