package gbaMus;

import java.io.FileNotFoundException;
import java.io.RandomAccessFile;
import java.io.FileOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;

/*
 * GBA Sound Font Riper (c) 2012 by Bregalad
 * This is free and open source software.
 * 
 * This java programs extracts soundfont data from a GBA game using
 * Nintendo's "sappy" engine (which ~90% of commercial GBA games are using),
 * and converts it to the widely used SF2 Sound Font format.
 */

public class SoundFontRiper {
	static RandomAccessFile inGBA;
	static DataOutputStream outSF2;
	static DataOutputStream outTxt;
	static RandomAccessFile psg_data;

	static boolean verbose_flag = false;
	static boolean verbose_output_to_file = false;
	static boolean change_sample_rate = false;
	static boolean psg_file_found = false;
	static boolean gm_preset_names = false;

	static int sample_rate = 22050;
	static ArrayList<Integer> addresses;
	static int current_address;
	static int current_bank;
	static int current_instrument;
	static int main_volume = 15;

	static SF2 sf2;

	//General MIDI instrument names
	static final String[] GM_Instruments = new String[] {
		"Acoustic Grand Piano",	"Bright Acoustic Piano", "Electric Grand Piano", "Honky-tonk Piano", "Rhodes Piano", "Chorused Piano",
		"Harpsichord",	"Clavinet",	"Celesta", "Glockenspiel", "Music Box",	"Vibraphone", "Marimba", "Xylophone", "Tubular Bells",	"Dulcimer",
		"Hammond Organ", "Percussive Organ", "Rock Organ", "Church Organ", "Reed Organ", "Accordion", "Harmonica", "Tango Accordion",
		"Acoustic Guitar (nylon)", "Acoustic Guitar (steel)", "Electric Guitar (jazz)", "Electric Guitar (clean)", "Electric Guitar (muted)",
		"Overdriven Guitar", "Distortion Guitar", "Guitar Harmonics", "Acoustic Bass", "Electric Bass (finger)", "Electric Bass (pick)",
		"Fretless Bass", "Slap Bass 1", "Slap Bass 2", "Synth Bass 1", "Synth Bass 2", "Violin", "Viola", "Cello", "Contrabass",
		"Tremelo Strings", "Pizzicato Strings", "Orchestral Harp", "Timpani", "String Ensemble 1", "String Ensemble 2",	"SynthStrings 1",
		"SynthStrings 2", "Choir Aahs", "Voice Oohs", "Synth Voice", "Orchestra Hit", "Trumpet", "Trombone", "Tuba", "Muted Trumpet",
		"French Horn", "Brass Section", "Synth Brass 1", "Synth Brass 2", "Soprano Sax", "Alto Sax", "Tenor Sax", "Baritone Sax",
		"Oboe", "English Horn", "Bassoon", "Clarinet", "Piccolo", "Flute", "Recorder", "Pan Flute", "Bottle Blow", "Shakuhachi", "Whistle",
		"Ocarina", "Lead 1 (square)", "Lead 2 (sawtooth)", "Lead 3 (calliope lead)", "Lead 4 (chiff lead)", "Lead 5 (charang)",
		"Lead 6 (voice)", "Lead 7 (fifths)", "Lead 8 (bass + lead)", "Pad 1 (new age)", "Pad 2 (warm)", "Pad 3 (polysynth)", "Pad 4 (choir)",
		"Pad 5 (bowed)", "Pad 6 (metallic)", "Pad 7 (halo)", "Pad 8 (sweep)", "FX 1 (rain)", "FX 2 (soundtrack)", "FX 3 (crystal)",
		"FX 4 (atmosphere)", "FX 5 (brightness)", "FX 6 (goblins)",	"FX 7 (echoes)", "FX 8 (sci-fi)", "Sitar", "Banjo", "Shamisen", "Koto",
		"Kalimba", "Bagpipe", "Fiddle", "Shanai", "Tinkle Bell", "Agogo", "Steel Drums", "Woodblock", "Taiko Drum", "Melodic Tom",
		"Synth Drum", "Reverse Cymbal", "Guitar Fret Noise", "Breath Noise", "Seashore", "Bird Tweet", "Telephone Ring", "Helicopter",
		"Applause",	"Gunshot"
	};

	public static void main(String[] args) throws IOException{
		System.out.println("GBA ROM sound font ripper (c) 2012 Bregalad\n");

		//Parse arguments
		addresses = new ArrayList<Integer>();
		parseArguments(args);


		//Create SF2 class
		sf2 = new SF2(sample_rate);

		//Attempt to access psg_data file
		try{
			psg_data = new RandomAccessFile("psg_data.bin", "r");
			psg_file_found = true;
		}catch (FileNotFoundException e) {
			System.out.println("psg_data.raw file not found !");
			System.out.println("PSG Instruments can't be dumped.");
		}

		//Read instrument data from input GBA file
		int[][] instrData = new int[128][3];

		//Decode all banks
		for(current_bank = 0; current_bank < addresses.size(); current_bank ++) {
			current_address = addresses.get(current_bank);
			//Seek at the start of the sound bank
			inGBA.seek(current_address);
			//Read entire sound bank in memory
			for(int i=0; i!=128; i++)
				for(int j=0; j!=3; j++)
					instrData[i][j] = inGBA.readInt();
	
			//Decode all instruments
			for(current_instrument = 0; current_instrument < 128; current_instrument ++, current_address += 12) {
				print("\nBank : " + current_bank + ", Instrument : " +
						current_instrument + " @0x" + Integer.toHexString(current_address));

				//Ignore unused instruments
				if(instrData[current_instrument][0] == 0x13c0000
				&& instrData[current_instrument][1] == 0x2000000
				&& instrData[current_instrument][2] == 0xF00) {
					print(" (unused)");
					continue;
				}

				if(verbose_flag)
					verboseInstrument(instrData[current_instrument], false);

				//Build equivalent SF2 instrument
				buildInstrument(instrData[current_instrument][0]>>>24);
			}
		}

		if(verbose_output_to_file) {
			print("\n\n EOF");
			outTxt.close();
		}

		System.out.print("\nDump complete, now outputing SF2 data...");

		sf2.write(outSF2);
		//Close files
		inGBA.close();
		System.out.println(" Done !");
	}

	//Convert a GBA instrument in its SF2 counterpart
	//if any kind of error happens, it will do nothing and exit
	static void buildInstrument(int instrType) {
		String name;
		if(gm_preset_names)
			name = GM_Instruments[current_instrument];
		else
			//(poetic) name of the SF2 preset...
			name = "Type " + instrType + " @0x" + Integer.toHexString(current_address);

		try{
			switch (instrType) {
			//Sampled instrument types
			case 0:
			case 8: {
				int i = sf2.instruments.buildSampledInstrument(inGBA, current_address);
				sf2.add_new_preset(name, current_instrument, current_bank);
				sf2.add_new_preset_bag();
				//Add initial attenuation preset to balance volume between sampled and GB instruments
				addAttenuationPreset();
				sf2.add_new_preset_generator(SFGenerator.instrument, i);
			}break;

			//GameBoy pulse wave instruments
			case 1:
			case 2:
			case 9:
			case 10: {
				//Can only convert them if the psg_data file is found
				if(psg_file_found) {
					int i = sf2.instruments.buildPulseInstrument(inGBA, psg_data, current_address);
					sf2.add_new_preset(name, current_instrument, current_bank);
					sf2.add_new_preset_bag();
					sf2.add_new_preset_generator(SFGenerator.instrument, i);
				}else {
					System.out.println("Warning : Couldn't convert GameBoy instrument because psg_data file is missing.");
				}
			}break;

			//GameBoy channel 3 instrument
			case 3:
			case 11: {
				int i = sf2.instruments.buildGB3Instrument(inGBA, current_address);
				sf2.add_new_preset(name, current_instrument, current_bank);
				sf2.add_new_preset_bag();
				sf2.add_new_preset_generator(SFGenerator.instrument, i);
			}break;

			//GameBoy noise instruments, not supported yet
			case 4:
			case 12 : {
				if(psg_file_found) {
					int i = sf2.instruments.buildNoiseInstrument(inGBA, psg_data, current_address);
					sf2.add_new_preset(name, current_instrument, current_bank);
					sf2.add_new_preset_bag();
					sf2.add_new_preset_generator(SFGenerator.instrument, i);
				}else {
					System.out.println("Warning : Couldn't convert GameBoy instrument because psg_data file is missing.");
				}
			}break;

			//Key split instrument
			case 0x40: {
				int i = sf2.instruments.buildKeySplitInstrument(inGBA, current_address);
				sf2.add_new_preset(name, current_instrument, current_bank);
				sf2.add_new_preset_bag();
				//Add initial attenuation preset to balance volume between sampled and GB instruments
				addAttenuationPreset();
				sf2.add_new_preset_generator(SFGenerator.instrument, i);
			}break;

			//Every key split instrument
			case 0x80: {
				int i = sf2.instruments.buildEveryKeySplitInstrument(inGBA, psg_data, current_address);
				sf2.add_new_preset(name, current_instrument, current_bank);
				sf2.add_new_preset_bag();
				//Add initial attenuation preset to balance volume between sampled and GB instruments
				addAttenuationPreset();
				sf2.add_new_preset_generator(SFGenerator.instrument, i);
			}break;

			//Other instrument types ?
			default:
			//	System.out.println("Warning : Unknown instrument type 0x" + Integer.toHexString(instrType) + ".");
			}

		//If there is any error in the process just ignore it and silently continue
		//In fact dozen of errors always happened all the times so I removed any form of error messages
		}catch (Exception e) {};
	}

	//Add initial attenuation preset to balance between GameBoy and sampled instruments
	static void addAttenuationPreset() {
		if(main_volume < 15) {
			final double attenuation = (short)(100.0 * Math.log(15.0/main_volume));
			sf2.add_new_preset_generator(SFGenerator.initialAttenuation, (short)attenuation);
		}
	}


	//This function read instrument data and outputs info on the screen or on the verbose file
	//it's not actually needed to convert the data to SF2 format, but is very useful for debugging
	static void verboseInstrument(int[] instr, boolean recursive) throws IOException {
		//Do nothing with unused instruments
		if(instr[0] == 0x13c0000 && instr[1] == 0x2000000 && instr[2] == 0xF00)
			return;

		int instrType = instr[0] >>> 24;
		print("  Type : 0x" + Integer.toHexString(instrType) + "  ");
		switch(instrType) {
		
			//Sampled instruments
			case 0 :
			case 8 : {
				int sadr = GBAMusRiper.getPointer(instr[1]);
				print("(sample @0x" + Integer.toHexString(sadr) + ")\n");

				try {
					inGBA.seek(sadr);
					int Loop = inGBA.readInt();
					int Pitch = Integer.reverseBytes(inGBA.readInt());
					int LoopPos = Integer.reverseBytes(inGBA.readInt());
					int Len = Integer.reverseBytes(inGBA.readInt());
					print("      Pitch : " + (double)Pitch/1024 + "\n");
					print("      Length : " + Len + "\n");

					if(Loop == 0)
						print("      Not looped\n");
					else if(Loop == 0x40)
						print("      Loop enabled at : " + LoopPos + "\n");
					else print("      Unknown loop type\n");
					adsr(instr[2]);
				} catch (IOException e) {
					print("Invalid instrument (an exception occured)\n");
				}
			}break;

			//Pulse channel 1 instruments
			case 1 :
			case 9 : {
				print("(GB pulse channel 1)\n");
				if((char)instr[0] != 8) {
					//Display sweep if enabled on GB channel 1
					print("      Sweep : " + (instr[0] & 0xFF) + "\n");
				}
				adsr(instr[2]);
				dutyCycle((char)(instr[1]>>24));
			}break;
			
			//Pulse channel 2 instruments
			case 2 :
			case 10 : {
				print("(GB pulse channel 2)\n");
				adsr(instr[2]);
				dutyCycle((char)(instr[1]>>24));
			}break;

			//Channel 3 instruments
			case 3 :
			case 11 : {
				print("(GB channel 3)\n");
				adsr(instr[2]);
				print("      Waveform : \n");

				try{
					//Seek to waveform's location
					inGBA.seek(GBAMusRiper.getPointer(instr[1]));
					int[] waveform = new int[32];

					for(int j=0; j<16; j++) {
						byte a = inGBA.readByte();
						waveform[2*j] = (a>>4) & 0x0F;
						waveform[2*j+1] = a & 0xF;
					}

					//Display waveform in text format
					for(int j=7; j>=0; j--) {
						for(int k=0; k!=32; k++) {
							if(waveform[k] == 2*j)
								print("_");
							else if(waveform[k] == 2*j+1)
								print("-");
							else
								print(" ");
						}
						print("\n");
					}
				}catch(IOException e) {
					print("Invalid instrument (an exception occured)\n");
				}
			}break;

			//Noise instruments
			case 4 :
			case 12 :
				print("(GB noise channel 4)\n");
				adsr(instr[2]);
				if(instr[1]  == 0)
					print("      long random sequence\n");
				else
					print("      short random sequence\n");
				break;

			//Key-split instruments
			case 0x40 :
				print("Key-split instrument\n");

				if(!recursive) {
					try{
					//seek to key table's location
						inGBA.seek(GBAMusRiper.getPointer(instr[2]));
						byte keys[] = new byte[128];
						boolean keys_used[] = new boolean[128];
						for(int k = 0; k!= 128; k++) {
							keys[k] = (byte)(inGBA.readByte() & 0x7F);
							keys_used[keys[k]] = true;
						}
	
						int instr_table = GBAMusRiper.getPointer(instr[1]);
						
						for(int k = 0; k!= 128; k++) {
							//Decode instruments used at least once in the key table
							if(keys_used[k]) {
								try{
									//Seek to the addressed instrument
									inGBA.seek(instr_table + 12*k);
									int[] sub_instr = new int[3];
									//Read the addressed instrument
									for(int l=0; l!=3; l++)
										sub_instr[l] = inGBA.readInt();
		
									print("\n      Sub_intrument " + k);
									verboseInstrument(sub_instr, true);
								}catch(IOException e) {
									print("Invalid sub-instrument (an exception occured)\n");
								}
							}
						}
					}catch (IOException e) {}
				} else
					print("   Illegal double-recursive instrument !\n");
				break;
			
			//Every key split instruments
			case 0x80 :
				print("Every key split instrument\n");
				
				if(!recursive) {
					int address = GBAMusRiper.getPointer(instr[1]);
					for(int k = 0; k<128; ++k) {
						try{
							inGBA.seek(address + k*12);
							int[] key_instr = new int[3];
							for(int l=0; l!=3; l++)
								key_instr[l] = inGBA.readInt();
							
							print("\n   Key " + k);
							verboseInstrument(key_instr, true);
						}catch(IOException e) {
							print("Illegal sub-instrument (an exception occured)\n");
						}
					}

				} else	//Prevent instruments with multiple recursivities
					print("   Illegal double-recursive instrument !\n");
				break;

			default :
				print("Unknown instrument type\n");
				break;
		}
		if(recursive) {
			print("      Key : " + ((instr[1] >> 16)&0xFF));
			print(", Pan : " + (instr[1] & 0xFF) + "\n");
		}
	}

	//Display ADSR values used
	static void adsr(int adsr) throws IOException {
		int Attack = (adsr >> 24)&0xFF;
		int Decay = (adsr >> 16)&0xFF;
		int Sustain = (adsr >> 8)&0xFF;
		int Release = adsr&0xFF;
		//Print ADSR values
		print("      ADSR : " + Attack + ", " + Decay + ", " + Sustain + ", " + Release + "\n");
	}

	//Display duty cycle used
	static void dutyCycle(char duty) throws IOException {
		final String[] cycles = {"12.5%", "25%", "50%", "75%"};
		print("      Duty cycle : " + cycles[duty&3] + "\n");
	}

	//Display verbose to console or output to file if requested
	static void print(String s) throws IOException {
		if(verbose_flag)
			if(!verbose_output_to_file)
				System.out.print(s);
			else
				outTxt.writeBytes(s);
	}
	
	static void parseArguments(String[] args) {
		if (args.length <3) {
			printInstructions();
		}

		try{
			inGBA = new RandomAccessFile(args[0], "r");			//Input File
		}catch(FileNotFoundException e) {
			System.out.println("Invalid input GBA file : " + args[0]);
			System.exit(-1);
		}

		try{
			String outsf2_name = args[1];
			if (!outsf2_name.contains(".") || !outsf2_name.substring(outsf2_name.lastIndexOf('.')).equals(".sf2")) 
				outsf2_name = outsf2_name + ".sf2";

			outSF2 = new DataOutputStream(new FileOutputStream(outsf2_name));
		}catch(FileNotFoundException e) {
			System.out.println("Invalid output file : " + args[1]);
			System.exit(-1);
		}


		for(int i = 2; i<args.length; i++) {

			//Enable verbose if -v flag encountered in arguments list
			if(args[i].substring(0,2).equals("-v")) {
				verbose_flag = true;
				//Verbose to file if a file name is given
				if(args[i].length() > 2) {
					verbose_output_to_file = true;

					try{
						outTxt = new DataOutputStream(new FileOutputStream(args[i].substring(2)));
					}catch(FileNotFoundException e) {
						System.out.println("Invalid output file : " + args[i].substring(2));
						System.exit(-1);
					}
				}
			}

			//Change sampling rate if -s is encountered
			else if(args[i].substring(0,2).equals("-s")) {
				if(args[i].length() <= 2) printInstructions();
				change_sample_rate = true;
				sample_rate = Integer.parseInt(args[i].substring(2));
			}

			//Change main volume if -mv is encountered
			else if(args[i].substring(0,3).equals("-mv")) {
				if(args[i].length()<4) printInstructions();

				int volume = Integer.parseInt(args[i].substring(3));
				if(volume<1 || volume>15) printInstructions();
				main_volume = volume;
			}
			
			else if(args[i].equals("-gm")) {
				gm_preset_names = true;
			}

			//Add address to list
			else if(args[i].substring(0,2).equals("0x")) {
				if(args[i].length() <= 2) printInstructions();
				//Add addresses
				addresses.add(Integer.parseInt(args[i].substring(2), 16));				
			}
			else printInstructions();
		}
	}

	static void printInstructions() {
		System.out.println("Dumps a sound bank (or a list of sound banks) from a GBA game which is using the sappy sound engine to SoundFont 2.0 (.sf2) format");
		System.out.println("Usage : SoundFontRipper in.gba out.sf2 [-v[Out.txt]] [-s22050] address1 [address2] ....");
		System.out.println("instruments at address1 will be dumped to Bank0, instruments at address2 to Bank1, etc.....");
		System.out.println("addresses should be in hex format. Example : \"0x123abc\"");
		System.out.print("-v : verbose : Display info about the sound font in text format. If -v is followed by a file name," +
						"info is output to the specified file instead.");
		System.out.println("-s : Sampling rate for samples. Default : 22050 Hz");
		System.out.println("-gm : Give General MIDI names to presets. Note that this will only change the names and will NOT " +
							"magically turn the soundfont into a General MIDI compliant soundfont.");
		System.out.println("-mv : Main volume for sample instruments. Range : 1-15. Game Boy channels are unnaffected.");
		System.exit(-1);
	}
}

